#include "flipdot/flipdot.hh"

#include <array>
#include <cmath>
#include <cstddef>
#include <cstdint>
#include <sstream>

#include <boost/chrono.hpp>
#include <boost/thread.hpp>

namespace {
  flipdot::image xayax_presents() {
    flipdot::image img(0);

    flipdot::render_text(img, 15,  4, "XayaX");
    flipdot::render_text(img,  6, 20, "presents");

    return img;
  }

  flipdot::image flippin_the_dots() {
    flipdot::image img(0);

    flipdot::render_text(img, 6,  4, "Flippin'");
    flipdot::render_text(img, 6, 20, "The Dots");

    return img;
  }

  class plasma_driver {
  public:
    // Simple sinus plasma. Algorithm found on Wikipedia somewhere (i.e., also blatantly stolen).
    plasma_driver(flipdot::display &disp, double t_init)
      : display_(disp),
        t(t_init)
      {
        for (std::size_t i = 0; i < sintable.size(); ++i) {
          sintable[i] = std::min(255, int(128 + 127 * std::sin(i * 3.1415926 * 2 / sintable.size())));
        }

        // for the color palette, run once around the HSB color space and end up where we
        // started; thus we get smooth color transitions everywhere.
        for(std::size_t i = 0; i < palette.size(); ++i) {
          palette[i] = (i & 0x1f) > 0x0f;
        }
      }

    void tick() {
      int y, x;
      std::size_t i;

      decltype(wave) speed_x;
      decltype(wave) speed_y;
      decltype(wave) workwave_x;
      decltype(wave) workwave_y;

      flipdot::image img;
      flipdot::image img_text;

      t += 0.02;

      if(t < 24.8) {
        img_text = xayax_presents();
      } else if(t < 26.79) {
        img_text = flippin_the_dots();
      }

      // Speed-Werte der Wellen vorberechnen
      speed_x[0] = 3000 * std::sin ( t * 1.0 + 1 * 1.5);
      speed_y[0] = 3000 * std::cos ( t * 1.1 + 2 * 1.5);
      speed_x[1] =  700 * std::sin (-t * 1.2 + 3 * 1.5);
      speed_y[1] =  700 * std::cos (-t * 1.3 + 4 * 1.5);
      speed_x[2] = 2300 * std::sin ( t * 1.4 + 5 * 1.5);
      speed_y[2] = 2300 * std::cos ( t * 1.5 + 6 * 1.5);
      speed_x[3] = 1700 * std::sin ( t * 1.6 + 7 * 1.5);
      speed_y[3] = 1700 * std::cos ( t * 1.7 + 8 * 1.5);

      for(i = 0; i < wave.size(); ++i) {
        workwave_y[i] = wave[i] += speed_x[i];
      }

      for(y = 0; y < img.height(); ++y) {
        for(i = 0; i < wave.size(); ++i) {
          workwave_x[i] = workwave_y[i];
        }

        for(x = 0; x < img.width(); ++x) {
          img.pixel_set(x, y, palette[(sintable[(workwave_x[0] >> 8) & 1023] +
                                       sintable[(workwave_x[1] >> 8) & 1023] +
                                       sintable[(workwave_x[2] >> 8) & 1023] +
                                       sintable[(workwave_x[3] >> 8) & 1023]) >> 2]);

          for (i = 0; i < wave.size(); ++i) {
            workwave_x[i] += speed_x[i];
          }
        }

        for (i = 0; i < wave.size(); ++i) {
          workwave_y[i] += speed_y[i];
        }
      }

      img ^= img_text;
      img.flush(display_);

//      std::cout << t << '\r' << std::flush;
    }

    double get_t() const { return t; }

  private:
    flipdot::display &display_;
    double t = 0.0;
    std::array<std::uint8_t, 1024> sintable;
    std::array<int, 4> wave { { 0 } };
    std::array<int, 256> palette;
  };

  void plasma(flipdot::display &disp, double t_init, double t_end) {
    //boost::chrono::milliseconds const interval(65);
    boost::posix_time::milliseconds const interval(75);

    plasma_driver p(disp, t_init);

//  auto now  = boost::chrono::system_clock::now();
    auto now  = boost::get_system_time();
    auto next = now + interval;

    while(p.get_t() < t_end) {
      p.tick();

      now  = next;
      next = now + interval;

      // boost::this_thread::sleep_until(now);
      boost::this_thread::sleep(now);
    }
  }
}

void part_plasma(flipdot::display &disp) {
  double t_init = 23.0;
  double t_end  = 32.89;

  plasma(disp, t_init, t_end);

  boost::this_thread::sleep(boost::posix_time::milliseconds(500));
}
