#include "point.hh"
#include "vec2d.hh"

#include <flipdot/flipdot.hh>

#include <boost/chrono.hpp>
#include <boost/thread.hpp>

#include <algorithm>
#include <cmath>
#include <functional>
#include <iostream>
#include <stdexcept>
#include <vector>

namespace {
  double square(double x) { return x * x; }

  class junction {
  public:
    junction(double x, double y)
      : pos_     (x, y),
        pos_last_(x, y),
        accel_   (0, 0)
      { }

    point const &pos     () const { return pos_     ; }
    point const &pos_prev() const { return pos_last_; }
    vec2d const &accel   () const { return accel_   ; }

    void accelerate(vec2d const &v) {
      accel_ = accel_ + v;
    }

    void integrate(double dt, double gravity) {
      double dt_squared = square(dt);
      double grav_y = -gravity * dt_squared;

      point tmp = pos_;

      ::move_rel(pos_, pos() - pos_prev() + accel() * dt_squared + vec2d(0.0, grav_y));
      pos_last_ = tmp;
      accel_    = vec2d(0.0, 0.0);
    }

    void move_to(point const &p) {
      pos_ = p;
    }

    void move_rel(vec2d const &v) {
      pos_ += v;
    }

    bool constrain(double left, double right, double bottom) {
      if      (pos_.x() < left ) pos_.x(left );
      else if (pos_.x() > right) pos_.x(right);

      if (pos_.y() < bottom) {
        pos_     .y(bottom);
        pos_last_.x(pos_.x());
        return true;
      }

      return false;
    }

  private:
    point pos_;
    point pos_last_;
    vec2d accel_;
  };

  class box {
  public:
    box(point const &topleft, vec2d const &extents)
      : topleft_(topleft), extents_(extents) { }

    double width () const { return extents_.x(); }
    double height() const { return extents_.y(); }

    double left  () const { return topleft_.x()           ; }
    double right () const { return topleft_.x() + width (); }
    double top   () const { return topleft_.y()           ; }
    double bottom() const { return topleft_.y() - height(); }

    point center() const { return point((left() + right()) / 2, (top() + bottom()) / 2); }

    bool contains(point const &p) const {
      return p.x() < right() && p.x() > left() && p.y() < top() && p.y() > bottom();
    }

    bool move_outside(junction &j) const {
      if(!contains(j.pos())) return false;

      vec2d  vd = j.pos_prev() - j.pos();
      double cx = vd.x() < 0.0 ? left  () : right ();
      double cy = vd.y() < 0.0 ? bottom() : top   ();

      vec2d vx = j.pos() + std::abs(j.pos().x() - cx) / length(vd) * vd;

      if(vx.y() < top() && vx.y() > bottom()) {
        j.move_to(vec2d(cx, j.pos().y()));
      } else {
        j.move_to(vec2d(j.pos().x(), cy));
        return cy == top();
      }

      return false;
    }

  private:
    point topleft_;
    vec2d extents_;
  };

  class blob {
  public:
    blob(double x, double y, double r, std::size_t n_parts)
      : p_   (),
        diff_(n_parts),
        n_   (n_parts) {
      if(n_parts < 5) {
        throw std::invalid_argument("How am I supposed to make a blob with no junctions?");
      }

      double cx = x;
      double cy = y;

      p_.reserve(n_parts);

      for(std::size_t i = 0; i < n_parts; ++i) {
        double ang = i * M_PI * 2 / n_parts;

        p_.emplace_back(cx + std::sin(ang) * r,
                        cy + std::cos(ang) * r);
      }

      side_length_ = distance(p_[0].pos(), p_[1].pos());
      target_area_ = area();
    }

    double area() {
      auto area_term = [](junction const &j1, junction const &j2) {
        return j1.pos().x() * j2.pos().y() - j2.pos().x() * j1.pos().y();
      };

      double a = area_term(p_.back(), p_.front());

      for(std::size_t i = 0; i + 1 < p_.size(); ++i){
        a += area_term(p_[i], p_[i + 1]);
      }

      a *= .5;

      return a;
    }

    void integrate(double dt) {
      for(auto &p : p_) {
        p.integrate(dt, gravity_force_);
      }
    }

    void fix_perimeter() {
      // Fix up side lengths

      for(int j = 0; j < perim_iters; ++j) {
        std::fill(begin(diff_), end(diff_), vec2d(0.0, 0.0));

        for(std::size_t i = 0; i < p_.size(); ++i) {
          std::size_t next = (i == p_.size() - 1) ? 0 : i + 1;

          vec2d vecd = p_[next].pos() - p_[i].pos();

          double distance = length(vecd);
          if (distance < EPS) distance = 1.0;

          double diffRatio = 1.0 - side_length_ / distance;

          vecd *= .5 * relaxFactor * diffRatio;

          diff_[i]    += vecd;
          diff_[next] -= vecd;
        }

        for(std::size_t i = 0; i < p_.size(); ++i) {
          p_[i].move_rel(diff_[i]);
        }
      }
    }

    void constrain_edges() {
      fix_perimeter();

      double perimeter = 0.0;

      std::fill(begin(n_), end(n_), vec2d());

      for(std::size_t i = 0; i < p_.size(); ++i) {
        std::size_t next = (i == p_.size() - 1) ? 0: i + 1;

        vec2d vecd = p_[next].pos() - p_[i].pos();
        double distance = length(vecd);

        if(distance < EPS) {
          distance = 1.0;
        }

        n_[i] = normal(vecd) / distance;
        perimeter += distance;
      }

      double deltaArea = target_area_ - area();
      double toExtrude = 0.5 * deltaArea / perimeter;

      for (std::size_t i = 0; i < p_.size(); ++i) {
        std::size_t next = (i == p_.size() - 1) ? 0 : i + 1;
        p_[next].move_rel(toExtrude * (n_[i] + n_[next]));
      }
    }

    void collide_walls() {
      for(auto &p : p_) {
        if(p.constrain(wall_left_, wall_right_, floor_y_)) {
          hit_floor = true;
        }
      }
    }

    int map_x(flipdot::image &img, double d) { return static_cast<int>(               ((d - wall_left_) / (wall_right_ - wall_left_)) * img.width () - 0.5); }
    int map_y(flipdot::image &img, double d) { return static_cast<int>(img.height() - ((d - floor_y_  ) / (ceil_y_     - floor_y_  )) * img.height() - 0.5); }

    void draw(flipdot::image &img) {
      draw_line(img,
                map_x(img, p_.back ().pos().x()), map_y(img, p_.back ().pos().y()),
                map_x(img, p_.front().pos().x()), map_y(img, p_.front().pos().y()));

      for(std::size_t i = 0; i < p_.size() - 1; ++i) {
        draw_line(img,
                  map_x(img, p_[i    ].pos().x()), map_y(img, p_[i    ].pos().y()),
                  map_x(img, p_[i + 1].pos().x()), map_y(img, p_[i + 1].pos().y()));
      }

      for(auto const &b : boxes_) {
        draw_box(img, b);
      }
    }

    void draw_box(flipdot::image &img, box const &b) {
      draw_line(img,
                map_x(img, b.left ()), map_y(img, b.top   ()),
                map_x(img, b.right()), map_y(img, b.top   ()));
      draw_line(img,
                map_x(img, b.right()), map_y(img, b.top   ()),
                map_x(img, b.right()), map_y(img, b.bottom()));
      draw_line(img,
                map_x(img, b.right()), map_y(img, b.bottom()),
                map_x(img, b.left ()), map_y(img, b.bottom()));
      draw_line(img,
                map_x(img, b.left ()), map_y(img, b.bottom()),
                map_x(img, b.left ()), map_y(img, b.top   ()));
    }

    enum {
      KEY_LEFT  = 1,
      KEY_RIGHT = 2,
      KEY_JUMP  = 4
    };

    void respond_to_input(unsigned keystate) {
      int lr = 0;
      bool jump = false;

      if (keystate & KEY_LEFT ) lr -= 1;
      if (keystate & KEY_RIGHT) lr += 1;
      if (keystate & KEY_JUMP ) jump = true;

      if(lr != 0) {
        for(std::size_t i = 0; i < p_.size(); ++i) {
          int next = (i == p_.size() - 1) ? 0 : i + 1;

          vec2d vecd = p_[next].pos() - p_[i].pos();
          double distance = length(vecd);

          if(distance < EPS) {
            distance = 1.0;
          }

          p_[i].accelerate(lr * rotateAccel * vecd / distance);
        }
      }

      if (jump && hit_floor) {
        double cmx = 0.0;

        for(auto &p : p_) {
          cmx += p.pos().x();
        }

        cmx /= p_.size();

        for(auto &p : p_) {
          p.accelerate(vec2d((p.pos().x() - cmx) * jumpSquish, 0.0));
        }
      }
    }

    void collide_boxes() {
      for(auto const &b : boxes_) {
        for(auto &p : p_) {
          if(b.move_outside(p)) {
            hit_floor = true;
          }
        }
      }
    }

    void step(flipdot::image &img, unsigned input_mask) {
      hit_floor = false;

      for(int i = 0; i < 6; ++i) {
        respond_to_input(input_mask);
        integrate(1.0 / 60.0);
        constrain_edges();
        collide_walls();
        collide_boxes();
      }

      draw(img);
    }

  private:
    std::vector<junction> p_;
    std::vector<point> diff_;
    std::vector<point> n_;

    std::vector<box> boxes_ {
      box({ -40.0,  10.0 }, { 20.0, 20.0 }),
        box({  20.0, -20.0 }, { 40.0, 10.0 }),
        box({  30.0,  20.0 }, { 10.0, 10.0 })
        };

    double side_length_;
    double target_area_;

    double wall_left_  = -60.0;
    double wall_right_ =  60.0;

    double floor_y_    = -30.0;
    double ceil_y_     =  30.0;

    double gravity_force_ = 9.8;

    int perim_iters = 5;
    double EPS = 0.0001;
    double relaxFactor = .9;
    double rotateAccel = 10.0;
    double jumpSquish = 40.0;

    bool hit_floor = false;
  };
}

void part_squishy(flipdot::display &disp) {
  boost::posix_time::milliseconds interval(40);
  auto now = boost::get_system_time();
  auto next = now + interval;

  std::vector<std::pair<unsigned, unsigned>> inputs {
    {  75, 0 },
    { 125, blob::KEY_LEFT },
    { 180, blob::KEY_RIGHT },
    { 190, blob::KEY_RIGHT | blob::KEY_JUMP },
    { 230, blob::KEY_RIGHT },
    { 250, blob::KEY_LEFT  | blob::KEY_JUMP },
    { 260, 0 },
    { 270, blob::KEY_JUMP },
    { 280, blob::KEY_RIGHT | blob::KEY_JUMP },
    { 330, blob::KEY_RIGHT },
    { 360, blob::KEY_LEFT  },
    { 370, blob::KEY_LEFT  | blob::KEY_JUMP },
    { 400, 0 },
    { 410, blob::KEY_LEFT  | blob::KEY_JUMP },
    { 500, blob::KEY_LEFT  },
    { 550, 0  },
    { 600, blob::KEY_RIGHT  },
    { 700, 0  },
    { 770, blob::KEY_LEFT },
    { 780, blob::KEY_LEFT | blob::KEY_JUMP  },
    { 800, blob::KEY_LEFT  },
    { 830, blob::KEY_RIGHT },
    { 840, blob::KEY_JUMP }
  };

  blob b(0.0, 0.0, 10.0, 40);

  auto input_iter = inputs.begin();

  for(unsigned step = 0; step < 857; ++step) {
    flipdot::image img;

    b.step(img, input_iter == inputs.end() ? 0u : input_iter->second);
    img.flush(disp);

//    std::cout << step << ", " << (input_iter == inputs.end() ? 0u : input_iter->second) << std::endl;

    if(input_iter != inputs.end() && step == input_iter->first) {
      ++input_iter;
    }

    boost::this_thread::sleep(next);
    next += interval;
  }
}
