#include "image.hh"

#include <algorithm>
#include <cassert>
#include <climits>
#include <stdexcept>

namespace flipdot {
  image::image(int gap_width, int height_modules, int width_modules)
    : fb_           (height_modules * width_modules * flipdot::MODULE_COLS),
      width_        (flipdot::MODULE_COLS * width_modules ),
      height_       (flipdot::MODULE_ROWS * height_modules + gap_width * (height_modules - 1)),
      module_height_(gap_width + flipdot::MODULE_ROWS),
      stripe_count_ (height_modules)
  {
    if(width_modules < 0 || height_modules < 0) {
      throw std::invalid_argument("Image dimensions invalid.");
    }

    clear();
  }

  void image::pixel_black(int x, int y) {
    if(x < 0 || x >= width_ || y < 0 || y >= height_) {
      return;
    }

    fb_[y / module_height_ * width_ + x] |= 1 << (y % module_height_);
  }

  void image::pixel_white(int x, int y) {
    if(x < 0 || x >= width_ || y < 0 || y >= height_) {
      return;
    }

    fb_[y / module_height_ * width_ + x] &= ~(1 << (y % module_height_));
  }

  void image::pixel_set(int x, int y, int val) {
    if(val) {
      pixel_black(x, y);
    } else {
      pixel_white(x, y);
    }
  }

  void image::clear() {
    std::fill(begin(fb_), end(fb_), 0);
  }

  namespace {
    class sprite_column_iterator {
    public:
      sprite_column_iterator(sprite const &sp, unsigned xskip)
        : data_ptr(sp.data), left_in_byte(CHAR_BIT) {
        if(xskip != 0) {
          int skip_bits = (xskip * sp.height);

          data_ptr += skip_bits / CHAR_BIT;
          left_in_byte = CHAR_BIT - skip_bits % CHAR_BIT;
        }
      }

      std::uint32_t pop(unsigned height) {
        std::uint32_t column = 0;

        if(left_in_byte > height) {
          column = *data_ptr >> (CHAR_BIT - left_in_byte) & ((1 << height) - 1);
          left_in_byte -= height;
        } else {
          column = *data_ptr >> (CHAR_BIT - left_in_byte);

          unsigned shlen = left_in_byte;
          unsigned n;

          for(n = height - shlen; n >= CHAR_BIT; n -= CHAR_BIT) {
            column |= *++data_ptr << shlen;
            shlen += CHAR_BIT;
          }

          column |= (*++data_ptr & ((1 << n) - 1)) << shlen;
          left_in_byte = CHAR_BIT - n;
        }

        return column;
      }

    private:
      std::uint8_t const *data_ptr;
      unsigned left_in_byte;
    };
  }

  void image::blt(int x, int y, sprite const &sp, std::uint32_t flags) {
    if(y >=  height_ ||
       x >=  width_  ||
       y <= -sp.height ||
       x <= -sp.width) {
      // Nichts zu tun.
      return;
    }

    int y_major = y / module_height_ - (y % module_height_ < 0);
    int y_minor = (y - y_major * module_height_);

    int xmove;
    int xcursor;
    int xend;
    unsigned xskip;

    if(flags & BLT_MIRRORED) {
      xmove   = -1;
      xcursor = std::min(x + sp.width, width_) - 1;
      xend    = std::max(x, 0) - 1;
      xskip   = x + sp.width - xcursor - 1;
    } else {
      xmove   = 1;
      xcursor = std::max(x, 0);
      xend    = std::min(x + sp.width, width_);
      xskip   = xcursor - x;
    }

    sprite_column_iterator col_iter(sp, xskip);

    while(xcursor != xend) {
      std::uint32_t column = col_iter.pop(sp.height);

      assert(xcursor >= 0);
      assert(xcursor < width_);

      if(y_major >= 0) {
        fb_[y_major * width_ + xcursor] |= column << y_minor;
      }

      unsigned stripe_max = std::min((sp.height + y_minor - 1) / module_height_,
                                     stripe_count_ - y_major - 1);

      for(unsigned stripe_ptr = std::max(1, -y_major); stripe_ptr <= stripe_max; ++stripe_ptr) {
        fb_[(y_major + stripe_ptr) * width_ + xcursor] |= column >> (stripe_ptr * module_height_ - y_minor);
      }

      xcursor += xmove;
    }
  }

  image &image::operator^=(image const &other) {
    for(std::size_t i = 0; i < std::min(fb_.size(), other.fb_.size()); ++i){
      fb_[i] ^= other.fb_[i];
    }

    return *this;
  }
}
