#include "gpio.hh"

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <fstream>
#include <stdexcept>
#include <sstream>
#include <string>

namespace beaglebone {
  namespace gpio {
    namespace {
      enum {
        START   = 0x44C00000,
        END     = 0x481AEFFF,
        CTRL    = 0x130,
        OE      = 0x134,
        DATAOUT = 0x13C,

        // S. AM335x ARM(R) Cortex(tm)-A8 Microprocessors (MPUs) Technical Reference Manual,
        // Sektion 2 (memory map), Seite 157, Links zu CM_PER (Sektion 8.1.12.1(.31-.33),
        // "CM_PER registers", für gpio1-3) und CM_WKUP (Sektion 8.1.12.2(.3), "CM_WKUP registers",
        // für gpio0)
        CLKCTRL_GPIO0 = 0x44E04008,
        CLKCTRL_GPIO1 = 0x44E000AC,
        CLKCTRL_GPIO2 = 0x44E000B0,
        CLKCTRL_GPIO3 = 0x44E000B4,

        CLKCTRL_MODENABLE = 0x2
      };

      class file_descriptor {
      public:
        file_descriptor(char const *fname, int mode)
          : fd_(open(fname, mode))
        {
          if(fd_ == -1) {
            throw std::logic_error("Could not open file.");
          }
        }

        ~file_descriptor() { close(fd_); }

        int descriptor() const { return fd_; }

      private:
        int fd_;
      };

      class memory_map {
      public:
        memory_map() : fd_("/dev/mem", O_RDWR), ptr_(0) {
          void *memmap = mmap(NULL,
                              END - START,
                              PROT_READ | PROT_WRITE,
                              MAP_SHARED,
                              fd_.descriptor(),
                              START);

          if(memmap == MAP_FAILED) {
            throw std::logic_error("Unable to map /dev/mem");
          }

          ptr_ = static_cast<std::uint32_t*>(memmap);

          // Clock Modules für GPIO-Ports anschalten.
          ptr_[(CLKCTRL_GPIO0 - START) / 4] = CLKCTRL_MODENABLE;
          ptr_[(CLKCTRL_GPIO1 - START) / 4] = CLKCTRL_MODENABLE;
          ptr_[(CLKCTRL_GPIO2 - START) / 4] = CLKCTRL_MODENABLE;
          ptr_[(CLKCTRL_GPIO3 - START) / 4] = CLKCTRL_MODENABLE;
        }

        ~memory_map() {
          munmap(const_cast<std::uint32_t*>(ptr_), END - START);
        }

        std::uint32_t volatile &reg(port_id p, unsigned offset) {
          return ptr_[(p - START + offset) / 4];
        }

        std::uint32_t volatile &ctrl(port_id p) { return reg(p, CTRL   ); }
        std::uint32_t volatile &port(port_id p) { return reg(p, DATAOUT); }
        std::uint32_t volatile &oe  (port_id p) { return reg(p, OE     ); }

      private:
        file_descriptor fd_;
        std::uint32_t volatile *ptr_;
      } gpio_map;
    }

    port::port(port_id id)
      : oe_      (gpio_map.oe  (id)),
        location_(gpio_map.port(id))
    { }
  }
}
