#include "iconverter.hh"

#include <flipdot/flipdot.hh>

#include <langinfo.h>
#include <form.h>

#include <boost/noncopyable.hpp>
#include <memory>

struct ncurses_init_handle {
  ncurses_init_handle() {
    char *lang = getenv("LANG");

    if(lang) {
      static_cast<void>(setlocale(LC_ALL, lang));
    }

    /* Initialize curses */
    initscr();
    cbreak();
    noecho();
    keypad(stdscr, TRUE);
  }

  ~ncurses_init_handle() {
    endwin();
  }
};

typedef std::unique_ptr<FIELD, int(*)(FIELD*)> field_handle;

field_handle make_field(int number) {
  field_handle result(new_field(1, 80, number * 2 + 1, 9, 0, 0), free_field);

  if(!result) {
    throw std::runtime_error("Failed to allocate ncurses field.");
  }

  set_field_back(result.get(), A_UNDERLINE);

  return result;
}

class text_form : private boost::noncopyable {
public:
  text_form()
    : field_objs_{ make_field(0),
                   make_field(1),
                   make_field(2),
                   make_field(3) },
      fields_{ field_objs_[0].get(),
               field_objs_[1].get(),
               field_objs_[2].get(),
               field_objs_[3].get(),
               nullptr },
      form_(new_form(fields_)) {
    if(form_ == nullptr) {
      throw std::runtime_error("Failed to allocate ncurses form");
    }

    form_opts_off(form_, O_BS_OVERLOAD);

    post_form(form_);

    for(int i = 0; i < 4; ++i) {
      mvprintw(i * 2 + 1, 0, "Zeile %d:", i + 1);
    }

    driver(REQ_FIRST_FIELD);

    refresh();
  }

  ~text_form() {
    unpost_form(form_);
    free_form(form_);
  }

  void driver(int c) {
    form_driver(form_, c);
  }

  void dump(flipdot::display &disp) {
    driver(REQ_PREV_PAGE);

    iconverter recode("ISO-8859-15", nl_langinfo(CODESET));

    flipdot::image img(0);

    for(int i = 0; i < 4; ++i) {
      flipdot::render_text(img, 3, i * 8, recode(field_buffer(fields_[i], 0)));
    }

    img.flush(disp);
  }

  void run(flipdot::display &disp) {
    int ch = KEY_F(1);
    bool immediate = true;

    do {
      switch(ch) {
      case KEY_DOWN:
      case '\t':
        driver(REQ_NEXT_FIELD);
        driver(REQ_END_LINE);
        break;

      case KEY_UP:
      case KEY_BTAB:
        driver(REQ_PREV_FIELD);
        driver(REQ_END_LINE);
        break;

      case KEY_LEFT:
        driver(REQ_LEFT_CHAR);
        break;

      case KEY_RIGHT:
        driver(REQ_RIGHT_CHAR);
        break;

      case KEY_HOME:
        driver(REQ_BEG_LINE);
        break;

      case KEY_END:
        driver(REQ_END_LINE);
        break;

      case 127: // Backspace
        driver(REQ_DEL_PREV);
        if(immediate) {
          dump(disp);
        }
        break;

      case KEY_DC:
        driver(REQ_DEL_CHAR);
        if(immediate) {
          dump(disp);
        }
        break;

      case KEY_F(5):
      case '\n':
        dump(disp);
        break;

      case KEY_F(1):
        if(immediate) {
          immediate = false;
          mvprintw(20, 0, "Ausgabe mit Return");
        } else {
          immediate = true;
          mvprintw(20, 0, "Sofortausgabe     ");
          dump(disp);
        }
        mvprintw(21, 0, "F1 zum Wechseln");
        driver(REQ_PREV_PAGE);

        break;

      default:
//        mvprintw(40, 0, "%o %d", ch, ch);
        driver(ch);
        if(immediate) {
          dump(disp);
        }
        break;
      }

      ch = getch();
    } while(ch != KEY_F(10) && ch != 0x18);
  }

private:
  field_handle field_objs_[4];
  FIELD      *fields_[5];
  FORM       *form_;
};

int main() {
  ncurses_init_handle ncurses;
  text_form frm;
  flipdot::display disp;

  frm.run(disp);
}
