#include "display.hh"

#include <unistd.h>

#include <boost/chrono.hpp>

#include <algorithm>
#include <csignal>
#include <functional>
#include <iostream>
#include <stdexcept>

namespace flipdot {
  namespace {
    inline uint16_t rotate_left(uint16_t x) { return x != 0x8000 ? x << 1 : 1; }
  }

  display::watchdog_process::~watchdog_process() {
    // in parent process, if watchdog is no longer required, kill it with fire.
    // FIXME: This is kind of ugly, but I see no simple way to make it clean and easy to
    // use. We could split the watchdog from the display and have the application coder
    // sort it out, but that's not a terribly nice API proposition.
    if(pid_ != 0) {
      kill(pid_, SIGKILL);
      close(pipes_[1]);
    }
  }

  void display::watchdog_process::start() {
    if(pid_ != 0) {
      throw std::logic_error("Tried to fork twice on the same flipdot::display::watchdog_process instance.");
    }

    if(0 != pipe(pipes_)) {
      throw std::runtime_error("Could not open pipes for watchdog process; error " + std::to_string(errno));
    }

    pid_ = fork();

    if(child_side()) {
      // IGNORE ALL THE THINGS!!1! Except SIGKILL, because that's what parent
      // sends us when we're no longer needed (and we can't ignore it anyway).
      sigset_t sigs;

      sigfillset(&sigs);
      sigdelset(&sigs, SIGKILL);
      sigprocmask(SIG_SETMASK, &sigs, nullptr);

      // The pipe will close upon parent death, then read will return
      // with eof. Checking for EINTR because paranoia is sanity.
      close(pipes_[1]);
      do {
        char c;
        errno = 0;

        // The return value of read does not matter to us here -- we're not trying to transfer data
        // but monitor if the other side still lives -- but gcc/glibc are annoyed with us if we don't
        // use it. Interestingly, casting to void as usual does not cut it here because read is declared
        // __attribute__((warn_unused_result)) as of glibc 2.19.
        ssize_t make_compiler_happy = read(pipes_[0], &c, 1);
        static_cast<void>(make_compiler_happy);
      } while(errno == EINTR);
    }

    close(pipes_[0]);
  }

  bool display::watchdog_process::child_side() const {
    return pid_ == 0;
  }

  display::display(std::size_t           module_count      ,
                   unsigned              delay_flip        ,
                   unsigned              delay_color       ,
                   unsigned              keyline_pause     ,
                   beaglebone::gpio::pin pin_coldata_data  ,
                   beaglebone::gpio::pin pin_strobe        ,
                   beaglebone::gpio::pin pin_oe_black      ,
                   beaglebone::gpio::pin pin_oe_white      ,
                   beaglebone::gpio::pin pin_rowselect_clk ,
                   beaglebone::gpio::pin pin_coldata_clk   ,
                   beaglebone::gpio::pin pin_rowselect_data)
    : sreg_rowselect(pin_rowselect_data, pin_rowselect_clk, MODULE_ROWS),
      sreg_coldata  (pin_coldata_data  , pin_coldata_clk  , module_count * (MODULE_COLS + MODULE_GAP)),

      pin_strobe  (pin_strobe),
      pin_oe_black(pin_oe_black),
      pin_oe_white(pin_oe_white),

      module_count(module_count),
      display_cols(module_count * MODULE_COLS),

      delay_flip (delay_flip),
      delay_color(delay_color),

      keyline_pause(keyline_pause),
      keyline_pause_left(keyline_pause),

      fb_current (display_cols, 0x0000),
      fb_physical(display_cols, 0xffff)
  {
    pin_neutralize();

    watchdog.start();

    if(watchdog.child_side()) {
      // watchdog process. If parent dies unexpectedly, pulls pins into neutral state.
      // In the child process, watchdog.start() waited for parent to die.
      std::cerr << "Parent died; neutralizing pins." << std::endl;
      pin_neutralize();

      // FIXME: Is it possible to guarantee sane cleanup without returning to the main program?
      std::quick_exit(0);
    } else {
      worker_thread = boost::scoped_thread<>(boost::thread(&display::run, this));
      wait();
    }
  }

  display::~display() {
    wait();
    interrupted = true;
    worker_thread.join();
  }

  void display::pin_neutralize() {
    pin_strobe  .out ();
    pin_oe_black.out ();
    pin_oe_white.out ();

    pin_oe_black.up  ();
    pin_oe_white.up  ();
    pin_strobe  .down();
  }

  bool display::busy() const {
    boost::lock_guard<boost::recursive_mutex> lock(mtx_framebuffer);
    return dirty_rows() != 0;
  }

  void display::wait() const {
    boost::unique_lock<boost::recursive_mutex> lock(mtx_framebuffer);
    cond_clean.wait(lock, [this]{ return !busy(); });
  }

  void display::flush(framebuffer const &fb) {
    wait();
    flush_immediate(fb);
  }

  void display::flush_immediate(framebuffer const &fb) {
    std::uint16_t new_dirty_rows_white = 0;
    std::uint16_t new_dirty_rows_black = 0;

    boost::lock_guard<boost::recursive_mutex> lock(mtx_framebuffer);

    /* Mark rows that have changed in the new framebuffer. */
    std::size_t datacols = std::min(display_cols, fb.size());

    for(std::size_t i = 0; i < datacols; ++i) {
      new_dirty_rows_white |= (fb[i] ^ fb_current[i]) & ~fb[i];
      new_dirty_rows_black |= (fb[i] ^ fb_current[i]) &  fb[i];
    }

    fb_current = fb;
    dirty_rows_white |= new_dirty_rows_white;
    dirty_rows_black |= new_dirty_rows_black;
  }

  template<typename t_colbit,
           typename t_transform,
           typename t_physupdate>
  void display::sreg_data(std::uint16_t row_mask,
                          t_colbit      f_colbit,
                          t_transform   f_transform,
                          t_physupdate  f_physupdate) {
    unsigned index = display_cols;

    sreg_rowselect.set(row_mask);

    boost::lock_guard<boost::recursive_mutex> lock(mtx_framebuffer);

    for(int i = 0; i < module_count; ++i) {
      push_module_gap();

      for(int j = 0; j < MODULE_COLS; ++j) {
        --index;

        std::uint16_t mask = f_colbit(fb_physical[index],
                                      fb_current [index],
                                      row_mask);

        sreg_coldata.push_bit(f_transform(mask));
        f_physupdate(index, mask);
      }
    }

    strobe();
  }

  void display::run() {
    enum {
      STATE_WHITE_NEXT,
      STATE_BLACK_NEXT
    };

    int key_state = STATE_WHITE_NEXT;

    while(!interrupted) {
      if(busy()) {
        boost::lock_guard<boost::recursive_mutex> lock(mtx_framebuffer);

        next_line();

        if(scan_mask & dirty_rows_white) {
          sreg_data(scan_mask,
                    [](uint16_t col_old, uint16_t col_new, uint16_t row_mask) { return (col_new ^ col_old) & ~col_new & row_mask; },
                    [](uint16_t x) { return x ? 0 : 1; },
                    [this](unsigned index, uint16_t mask) { fb_physical[index] &= ~mask; });
          pin_oe_white.down();
          dirty_rows_white &= ~scan_mask;
        } else {
          sreg_data(scan_mask,
                    [](uint16_t col_old, uint16_t col_new, uint16_t row_mask) { return (col_new ^ col_old) &  col_new & row_mask; },
                    [](uint16_t x) { return x ? 1 : 0; },
                    [this](unsigned index, uint16_t mask) { fb_physical[index] |=  mask; });
          pin_oe_black.down();
          dirty_rows_black &= ~scan_mask;
        }

        next_line();
      } else {
        scan_mask = 0x0001;

        if(keyline_pause_left == 0) {
          keyline_pause_left = keyline_pause;

          boost::lock_guard<boost::recursive_mutex> lock(mtx_framebuffer);

          sreg_data(keyline_mask,
                    [](uint16_t col_old, uint16_t, uint16_t row_mask) { return col_old & row_mask; },
                    [](uint16_t x) { return x ? 1 : 0; },
                    [](unsigned, uint16_t) { });

          if(key_state == STATE_WHITE_NEXT) {
            pin_oe_white.down();
            key_state = STATE_BLACK_NEXT;
          } else {
            pin_oe_black.down();
            key_state = STATE_WHITE_NEXT;
            keyline_mask = rotate_left(keyline_mask);
          }
        } else {
          --keyline_pause_left;
        }
      }

      cond_clean.notify_one();
      boost::this_thread::sleep_for(boost::chrono::microseconds(delay_flip));

      pin_oe_black.up();
      pin_oe_white.up();

      boost::this_thread::sleep_for(boost::chrono::microseconds(delay_color));
    }
  }

  void display::strobe() {
    pin_strobe.up  ();
    pin_strobe.down();
  }

  void display::push_module_gap() {
    for(unsigned i = 0; i < MODULE_GAP; ++i) {
      sreg_coldata.push_bit(0);
    }
  }

  void display::next_line() {
    std::uint16_t d = dirty_rows();

    if(d == 0) {
      scan_mask = rotate_left(scan_mask);
    } else {
      std::uint16_t mask_lower = ~(scan_mask - 1) & 0xffff;
      if(d & mask_lower) {
        d &= mask_lower;
      }

      scan_mask = d ^ ((d - 1) & d);
    }
  }
}
