#ifndef INCLUDED_FLIPDOT_DISPLAY_HH
#define INCLUDED_FLIPDOT_DISPLAY_HH

#include <beaglebone/gpio.hh>
#include <hw/shift_register.hh>

#include <cstdint>
#include <vector>

#include <boost/atomic.hpp>
#include <boost/thread.hpp>
#include <boost/thread/scoped_thread.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <sys/types.h>
#include <unistd.h>

namespace flipdot {
  enum {
    MODULE_ROWS  = 16,
    MODULE_COLS  = 20,
    MODULE_GAP   =  4
  };

  typedef std::vector<std::uint16_t> framebuffer;

  class display {
  public:
    display(std::size_t           module_count       =    6,
            unsigned              delay_flip         = 1500,
            unsigned              delay_color        =  128,
            unsigned              keyline_pause      =  128,
            // Note: If you want to use custom pins, you will also need to change the device tree overlay in flipdot.dts.
            beaglebone::gpio::pin pin_coldata_data   = beaglebone::gpio::pin(beaglebone::gpio::PORT2,  2),
            beaglebone::gpio::pin pin_strobe         = beaglebone::gpio::pin(beaglebone::gpio::PORT1, 13),
            beaglebone::gpio::pin pin_oe_black       = beaglebone::gpio::pin(beaglebone::gpio::PORT1, 15),
            beaglebone::gpio::pin pin_oe_white       = beaglebone::gpio::pin(beaglebone::gpio::PORT2,  5),
            beaglebone::gpio::pin pin_rowselect_clk  = beaglebone::gpio::pin(beaglebone::gpio::PORT2,  3),
            beaglebone::gpio::pin pin_coldata_clk    = beaglebone::gpio::pin(beaglebone::gpio::PORT2,  4),
            beaglebone::gpio::pin pin_rowselect_data = beaglebone::gpio::pin(beaglebone::gpio::PORT1, 12)
            // RFU: beaglebone::gpio::pin pin_rowselect_data = beaglebone::gpio::pin(beaglebone::gpio::PORT1, 14)
      );
    ~display();

    bool busy() const;
    void wait() const;

    void flush          (framebuffer const &fb);
    void flush_immediate(framebuffer const &fb);

    void run();

    inline framebuffer empty_framebuffer() const { return framebuffer(display_cols, 0); }

  private:
    void strobe();
    void push_module_gap();
    void next_line();
    void pin_neutralize();

    inline uint16_t dirty_rows() const {
      boost::lock_guard<boost::recursive_mutex> lock(mtx_framebuffer);
      return dirty_rows_white | dirty_rows_black;
    }

    template<typename t_colbit,
             typename t_transform,
             typename t_physupdate>
    void sreg_data(std::uint16_t row_mask,
                   t_colbit      f_colbit,
                   t_transform   f_transform,
                   t_physupdate  f_physupdate);

    hw::shift_register    sreg_rowselect;
    hw::shift_register    sreg_coldata;
    beaglebone::gpio::pin pin_strobe;
    beaglebone::gpio::pin pin_oe_black;
    beaglebone::gpio::pin pin_oe_white;

    std::size_t           module_count;
    std::size_t           display_cols;

    unsigned              delay_flip;
    unsigned              delay_color;

    unsigned              keyline_pause;

    std::uint16_t         dirty_rows_white   = 0xffff;
    std::uint16_t         dirty_rows_black   = 0;

    std::uint16_t         scan_mask          = 0x0001;
    std::uint16_t         keyline_mask       = 0x0001;
    unsigned              keyline_pause_left;

    framebuffer           fb_current;
    framebuffer           fb_physical;

    // Hardware protection watchdog process. If the main process crashes or is
    // aborted, the watchdog will pull the flipdot pins into a neutral state.
    struct watchdog_process {
    public:
      ~watchdog_process();

      void start     ();
      bool child_side() const;

    private:
      pid_t pid_ = 0;
      int   pipes_[2] { 0, 0 };
    };

    // Order is potentially important for constructor thread safety (and to make
    // it generally harder to introduce bugs). In particular, it is nice to have
    // it guaranteed that worker_thread is destroyed before the things it uses,
    // and also before the watchdog process.
    watchdog_process                      watchdog;
    boost::atomic<bool>                   interrupted { false };
    mutable boost::recursive_mutex        mtx_framebuffer;
    mutable boost::condition_variable_any cond_clean;
    boost::scoped_thread<>                worker_thread;

    // Note: Best insert new member objects before watchdog unless you have good
    // reasons to do otherwise.
  };
}

#endif /* FLIPDOT_H */
